==> firestore-api-lib-1.1.8.tgz: package/FirestoreClient.ts <==
import * as admin from 'firebase-admin';
import Firestore = FirebaseFirestore.Firestore;
import CollectionReference = FirebaseFirestore.CollectionReference
import DocumentReference = FirebaseFirestore.DocumentReference;
import DocumentSnapshot = FirebaseFirestore.DocumentSnapshot;
import Transaction = FirebaseFirestore.Transaction;
import WriteResult = FirebaseFirestore.WriteResult;
import chunk from "lodash.chunk";
import WhereFilterOp = FirebaseFirestore.WhereFilterOp;
import Query = FirebaseFirestore.Query;

const MAX_BATCH_SIZE = 500
const MAX_WHERE_IN = 10

export class FirestoreClient {
    private readonly firestore: Firestore

    public constructor(firestore: Firestore) {
        this.firestore = firestore
    }

    public getCollection<T>(collectionName: string, optionalFields: string[] = [], versioned = true) : FirestoreCollection<T> {
        return new FirestoreCollection<T>(
            this.firestore.collection(collectionName) as CollectionReference<T>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public runTransaction<T>(updateFunction: (transaction: Transaction) => Promise<T>,transactionOptions?:{maxAttempts?: number}
    ): Promise<T> {
        return this.firestore.runTransaction<T>(updateFunction, transactionOptions)
    }
}

export class FirestoreCollection<T> {

    private readonly collectionReference: CollectionReference<T>;
    private readonly firestore : Firestore
    private readonly optionalFields: string[]
    private readonly versioned: boolean

    public constructor(collectionReference: CollectionReference<T>, firestore : Firestore, optionalFields: string[], versioned = true ) {
        this.collectionReference = collectionReference
        this.firestore = firestore
        this.optionalFields = optionalFields
        this.versioned = versioned
    }

    public async documentExists(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<boolean> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(false)
        }
        return Promise.resolve(true)
    }

    public async getDocument(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<T | undefined> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(undefined)
        }

        return Promise.resolve(FirestoreUtil.decode(documentSnapshot))
    }


    public getSubCollection<V>(documentId: string, collectionName: string, optionalFields: string[] = [], versioned =  true) : FirestoreCollection<V> {
        this.collectionReference.doc(documentId).collection(collectionName)
        return new FirestoreCollection<V>(
            this.collectionReference.doc(documentId).collection(collectionName) as CollectionReference<V>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public async setDocument(id: string, data: T) : Promise<WriteResult> {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(FirestoreUtil.datesToTimestamps(data))
        const docRef = this.collectionReference.doc(id);
        return docRef.set(data, {merge : true})
    }

    public async listDocuments() : Promise<T[]> {
        const docRefs = await this.collectionReference.listDocuments();
        if (docRefs !== undefined) {
            return Promise.all(docRefs.map(async doc => FirestoreUtil.decode(await doc.get())))
        }
        return Promise.reject()
    }

    public async addDocument(data: T) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(data)
        const docRef : DocumentReference<T> = await this.collectionReference.add(FirestoreUtil.datesToTimestamps(data));
        if (docRef === undefined) {
            return Promise.reject()
        }

        return this.getDocument(docRef)
    }

    public async updateDocument(id: string, updateData: any, transaction?: Transaction) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(updateData)
        const docRef = this.collectionReference.doc(id);
        transaction ? await transaction.update(docRef, FirestoreUtil.datesToTimestamps(updateData)) : await docRef.update(updateData)
        return this.getDocument(id)
    }

    public async deleteDocument(id: string) : Promise<WriteResult> {
        return this.collectionReference.doc(id).delete()
    }


    public setDocumentInTransaction(id: string|DocumentReference<T>|undefined, data: T, transaction: Transaction) : Transaction {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(data)
        let docRef
        if (!id) {
            docRef = this.collectionReference.doc()
        } else {
            docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        }
        return transaction.set(docRef, FirestoreUtil.datesToTimestamps(data), {merge : true})
    }

    public async executeQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Promise<T[]> {
        const querySnapshot = await this.collectionReference.where(fieldPath, opStr, value).get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public startQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Query<T> {
        return this.collectionReference.where(fieldPath, opStr, value)
    }

    public async getQueryResult(query: Query<T>) : Promise<T[]> {
        const querySnapshot = await query.get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public bulkSet(data: T[]): void {
         Promise.all(chunk(data, MAX_BATCH_SIZE).map(async dataRefs => {
            const writeBatch = this.firestore.batch();
            (dataRefs).map(value => {
                    this.deleteOptionalFields(value)
                    if (this.versioned) this.applyVersionIncrement(value)
                    const docRef = value.id ? this.collectionReference.doc(value.id) : this.collectionReference.doc()
                    writeBatch.set(
                        docRef,
                        FirestoreUtil.datesToTimestamps(value),
                        {merge: true}
                    )
                }
            )
            await writeBatch.commit()
        })).catch(reason => {
            return Promise.reject()
        })
    }

    public async getDocumentsById(ids: string[], transaction?: Transaction) : Promise<T[]> {
        let result : T[] = []
        await Promise.all(chunk(ids, MAX_WHERE_IN).map(async idsChunk => {
            const query = this.collectionReference.where(admin.firestore.FieldPath.documentId(), "in", idsChunk as string[])
            const queryResult = await (transaction ? transaction.get(query) : query.get())
            result = result.concat(queryResult.docs.map(value => FirestoreUtil.decode(value)))
        }))

        return result
    }

    private applyVersionIncrement(data : T) {
        (data as any).version = admin.firestore.FieldValue.increment(1)
    }

    private deleteOptionalFields(data : T) {
        this.optionalFields
            .filter(prop => data[prop] === undefined)
            .map(value => data[value] = admin.firestore.FieldValue.delete())
    }
}

export class FirestoreUtil {
    public static decode<T>(documentSnapshot : DocumentSnapshot<T>) : T {
        return  {
            ...this.timestampsToDates(documentSnapshot.data()!),
            created: documentSnapshot.createTime!.toDate(),
            lastModified: documentSnapshot.updateTime!.toDate(),
            id : documentSnapshot.id
        } as T
    }

    public static timestampsToDates<T>(data: T) : T {
        Object.keys(data).forEach(key => {
            if (data[key] && typeof data[key] === 'object' && (data[key]["_seconds"] !== undefined && data[key]["_nanoseconds"] !== undefined))
                data[key] = data[key].toDate();
            else if (data[key] && typeof data[key] === 'object') {
                this.timestampsToDates(data[key])
            }
        });
        return data;
    }

    public static datesToTimestamps<T>(data: T) : T {
        Object.keys(data).forEach(key => {

            if (data[key] && typeof data[key] === 'object' && data[key] instanceof Date)
                data[key] = admin.firestore.Timestamp.fromMillis(Date.parse(data[key].toString()))
            else if (data[key] && typeof data[key] === 'object') {
                this.datesToTimestamps(data[key])
            } else if (data[key] && typeof data[key] === 'string') {
                const parsedDate = this.parseIsoDate(data[key])
                if (parsedDate != undefined) {
                    data[key] = admin.firestore.Timestamp.fromMillis(Date.parse(data[key].toString()))
                }
            }
        });
        return data;
    }

    private static parseIsoDate(input: string): Date|undefined {
        const iso = /^(\d{4})(?:-?W(\d+)(?:-?(\d+)D?)?|(?:-(\d+))?-(\d+))(?:[T ](\d+):(\d+)(?::(\d+)(?:\.(\d+))?)?)?(?:Z(-?\d*))?$/;

        const parts = input.match(iso);

        if (parts == null) {
            return undefined;
        }

        const year = Number(parts[1]);
        let months
        let days

        if (typeof parts[2] != "undefined") {
            /* Convert weeks to days, months 0 */
            const weeks = Number(parts[2]) - 1;
            days  = Number(parts[3]);

            if (typeof days == "undefined") {
                days = 0;
            }

            days += weeks * 7;

            months = 0;
        }
        else {
            if (typeof parts[4] != "undefined") {
                months = Number(parts[4]) - 1;
            }
            else {
                /* it's an ordinal date... */
                months = 0;
            }

            days   = Number(parts[5]);
        }
        let minutes
        let hours
        let seconds      = 0;
        let milliseconds = 0;
        let fractional;
        if (typeof parts[6] != "undefined" &&
            typeof parts[7] != "undefined")
        {
            hours        = Number(parts[6]);
            minutes      = Number(parts[7]);

            if (typeof parts[8] != "undefined") {
                const seconds      = Number(parts[8]);

                if (typeof parts[9] != "undefined") {
                    fractional   = Number(parts[9]);
                    milliseconds = fractional / 100;
                }
                else {
                    milliseconds = 0
                }
            }
            else {
                seconds      = 0;
                milliseconds = 0;
            }
        }
        else {
            hours        = 0;
            minutes      = 0;
            seconds      = 0;
            fractional   = 0;
            milliseconds = 0;
        }

        if (typeof parts[10] != "undefined") {
            /* Timezone adjustment, offset the minutes appropriately */
            const localzone = -(new Date().getTimezoneOffset());
            const timezone  = parts[10] as unknown as number * 60;

            minutes = Number(minutes) + (timezone - localzone);
        }

        return new Date(year, months, days, hours, minutes, seconds, milliseconds);
    }
}

==> firestore-api-lib-1.1.9.tgz: package/FirestoreClient.ts <==
import * as admin from 'firebase-admin';
import Firestore = FirebaseFirestore.Firestore;
import CollectionReference = FirebaseFirestore.CollectionReference
import DocumentReference = FirebaseFirestore.DocumentReference;
import DocumentSnapshot = FirebaseFirestore.DocumentSnapshot;
import Transaction = FirebaseFirestore.Transaction;
import WriteResult = FirebaseFirestore.WriteResult;
import chunk from "lodash.chunk";
import WhereFilterOp = FirebaseFirestore.WhereFilterOp;
import Query = FirebaseFirestore.Query;

const MAX_BATCH_SIZE = 500
const MAX_WHERE_IN = 10

export class FirestoreClient {
    private readonly firestore: Firestore

    public constructor(firestore: Firestore) {
        this.firestore = firestore
    }

    public getCollection<T>(collectionName: string, optionalFields: string[] = [], versioned = true) : FirestoreCollection<T> {
        return new FirestoreCollection<T>(
            this.firestore.collection(collectionName) as CollectionReference<T>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public runTransaction<T>(updateFunction: (transaction: Transaction) => Promise<T>,transactionOptions?:{maxAttempts?: number}
    ): Promise<T> {
        return this.firestore.runTransaction<T>(updateFunction, transactionOptions)
    }
}

export class FirestoreCollection<T> {

    private readonly collectionReference: CollectionReference<T>;
    private readonly firestore : Firestore
    private readonly optionalFields: string[]
    private readonly versioned: boolean

    public constructor(collectionReference: CollectionReference<T>, firestore : Firestore, optionalFields: string[], versioned = true ) {
        this.collectionReference = collectionReference
        this.firestore = firestore
        this.optionalFields = optionalFields
        this.versioned = versioned
    }

    public async documentExists(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<boolean> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(false)
        }
        return Promise.resolve(true)
    }

    public async getDocument(id: string|DocumentReference<T>, transaction?: Transaction) : Promise<T | undefined> {
        const docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        const documentSnapshot = transaction ? await transaction.get(docRef) : await docRef.get()
        if (documentSnapshot === undefined || !documentSnapshot.exists) {
            return Promise.resolve(undefined)
        }

        return Promise.resolve(FirestoreUtil.decode(documentSnapshot))
    }


    public getSubCollection<V>(documentId: string, collectionName: string, optionalFields: string[] = [], versioned =  true) : FirestoreCollection<V> {
        this.collectionReference.doc(documentId).collection(collectionName)
        return new FirestoreCollection<V>(
            this.collectionReference.doc(documentId).collection(collectionName) as CollectionReference<V>,
            this.firestore,
            optionalFields,
            versioned
        )
    }

    public async setDocument(id: string, data: T) : Promise<WriteResult> {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(FirestoreUtil.datesToTimestamps(data))
        const docRef = this.collectionReference.doc(id);
        return docRef.set(data, {merge : true})
    }

    public async listDocuments() : Promise<T[]> {
        const docRefs = await this.collectionReference.listDocuments();
        if (docRefs !== undefined) {
            return Promise.all(docRefs.map(async doc => FirestoreUtil.decode(await doc.get())))
        }
        return Promise.reject()
    }

    public async addDocument(data: T) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(data)
        const docRef : DocumentReference<T> = await this.collectionReference.add(FirestoreUtil.datesToTimestamps(data));
        if (docRef === undefined) {
            return Promise.reject()
        }

        return this.getDocument(docRef)
    }

    public async updateDocument(id: string, updateData: any, transaction?: Transaction) : Promise<T | undefined> {
        if (this.versioned) this.applyVersionIncrement(updateData)
        const docRef = this.collectionReference.doc(id);
        transaction ? await transaction.update(docRef, FirestoreUtil.datesToTimestamps(updateData)) : await docRef.update(updateData)
        return this.getDocument(id)
    }

    public async deleteDocument(id: string) : Promise<WriteResult> {
        return this.collectionReference.doc(id).delete()
    }


    public setDocumentInTransaction(id: string|DocumentReference<T>|undefined, data: T, transaction: Transaction) : Transaction {
        this.deleteOptionalFields(data)
        if (this.versioned) this.applyVersionIncrement(data)
        let docRef
        if (!id) {
            docRef = this.collectionReference.doc()
        } else {
            docRef = typeof id === "string" ? this.collectionReference.doc(id) : id;
        }
        return transaction.set(docRef, FirestoreUtil.datesToTimestamps(data), {merge : true})
    }

    public async executeQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Promise<T[]> {
        const querySnapshot = await this.collectionReference.where(fieldPath, opStr, value).get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public startQuery(fieldPath: string, opStr: WhereFilterOp, value: any) : Query<T> {
        return this.collectionReference.where(fieldPath, opStr, value)
    }

    public async getQueryResult(query: Query<T>) : Promise<T[]> {
        const querySnapshot = await query.get()
        if (querySnapshot !== undefined) {
            return querySnapshot.docs.map(doc => FirestoreUtil.decode(doc))
        }
        return Promise.reject()
    }

    public bulkSet(data: T[]): void {
         Promise.all(chunk(data, MAX_BATCH_SIZE).map(async dataRefs => {
            const writeBatch = this.firestore.batch();
            (dataRefs).map(value => {
                    this.deleteOptionalFields(value)
                    if (this.versioned) this.applyVersionIncrement(value)
                    const docRef = value.id ? this.collectionReference.doc(value.id) : this.collectionReference.doc()
                    writeBatch.set(
                        docRef,
                        FirestoreUtil.datesToTimestamps(value),
                        {merge: true}
                    )
                }
            )
            await writeBatch.commit()
        })).catch(reason => {
            return Promise.reject()
        })
    }

    public async getDocumentsById(ids: string[], transaction?: Transaction) : Promise<T[]> {
        let result : T[] = []
        await Promise.all(chunk(ids, MAX_WHERE_IN).map(async idsChunk => {
            const query = this.collectionReference.where(admin.firestore.FieldPath.documentId(), "in", idsChunk as string[])
            const queryResult = await (transaction ? transaction.get(query) : query.get())
            result = result.concat(queryResult.docs.map(value => FirestoreUtil.decode(value)))
        }))

        return result
    }

    private applyVersionIncrement(data : T) {
        (data as any).version = admin.firestore.FieldValue.increment(1)
    }

    private deleteOptionalFields(data : T) {
        this.optionalFields
            .filter(prop => data[prop] === undefined)
            .map(value => data[value] = admin.firestore.FieldValue.delete())
    }
}

export class FirestoreUtil {
    public static decode<T>(documentSnapshot : DocumentSnapshot<T>) : T {
        return  {
            ...this.timestampsToDates(documentSnapshot.data()!),
            created: documentSnapshot.createTime!.toDate(),
            lastModified: documentSnapshot.updateTime!.toDate(),
            id : documentSnapshot.id
        } as T
    }

    public static timestampsToDates<T>(data: T) : T {
        Object.keys(data).forEach(key => {
            if (data[key] && typeof data[key] === 'object' && (data[key]["_seconds"] !== undefined && data[key]["_nanoseconds"] !== undefined))
                data[key] = data[key].toDate();
            else if (data[key] && typeof data[key] === 'object') {
                this.timestampsToDates(data[key])
            }
        });
        return data;
    }

    public static datesToTimestamps<T>(data: T) : T {
        Object.keys(data).forEach(key => {

            if (data[key] && typeof data[key] === 'object' && data[key] instanceof Date)
                data[key] = admin.firestore.Timestamp.fromDate(data[key])
            else if (data[key] && typeof data[key] === 'object') {
                this.datesToTimestamps(data[key])
            } else if (data[key] && typeof data[key] === 'string') {
                const parsedDate = this.parseIsoDate(data[key])
                if (parsedDate != undefined) {
                    data[key] = admin.firestore.Timestamp.fromDate(parsedDate)
                }
            }
        });
        return data;
    }

    private static parseIsoDate(input: string): Date|undefined {
        const iso = /^(\d{4})(?:-?W(\d+)(?:-?(\d+)D?)?|(?:-(\d+))?-(\d+))(?:[T ](\d+):(\d+)(?::(\d+)(?:\.(\d+))?)?)?(?:Z(-?\d*))?$/;

        const parts = input.match(iso);

        if (parts == null) {
            return undefined;
        }

        const year = Number(parts[1]);
        let months
        let days

        if (typeof parts[2] != "undefined") {
            /* Convert weeks to days, months 0 */
            const weeks = Number(parts[2]) - 1;
            days  = Number(parts[3]);

            if (typeof days == "undefined") {
                days = 0;
            }

            days += weeks * 7;

            months = 0;
        }
        else {
            if (typeof parts[4] != "undefined") {
                months = Number(parts[4]) - 1;
            }
            else {
                /* it's an ordinal date... */
                months = 0;
            }

            days   = Number(parts[5]);
        }
        let minutes
        let hours
        let seconds      = 0;
        let milliseconds = 0;
        let fractional;
        if (typeof parts[6] != "undefined" &&
            typeof parts[7] != "undefined")
        {
            hours        = Number(parts[6]);
            minutes      = Number(parts[7]);

            if (typeof parts[8] != "undefined") {
                const seconds      = Number(parts[8]);

                if (typeof parts[9] != "undefined") {
                    fractional   = Number(parts[9]);
                    milliseconds = fractional / 100;
                }
                else {
                    milliseconds = 0
                }
            }
            else {
                seconds      = 0;
                milliseconds = 0;
            }
        }
        else {
            hours        = 0;
            minutes      = 0;
            seconds      = 0;
            fractional   = 0;
            milliseconds = 0;
        }

        if (typeof parts[10] != "undefined") {
            /* Timezone adjustment, offset the minutes appropriately */
            const localzone = -(new Date().getTimezoneOffset());
            const timezone  = parts[10] as unknown as number * 60;

            minutes = Number(minutes) + (timezone - localzone);
        }

        return new Date(year, months, days, hours, minutes, seconds, milliseconds);
    }
}
